//==================================
// LOADPROF - Matt Pietrek 1994
// FILE: DUPFIXUP.C
//==================================
#include <windows.h>
#include <windowsx.h>
#include <stdlib.h>
#include <stdio.h>
#include "newexe.h"

int SortRelocs(const void * a, const void *b);  // Prototype the sort function

//
// Returns the number of duplicate fixups in an EXE or DLL
//
DWORD HasDuplicateFixups(LPSTR szModuleName)
{
    NEWEXEHEADER neheader = {0};
    DWORD neOffset;
    FILE * pFile;
    WORD i;
    DWORD cDuplicates = 0;

    // open the file
    pFile = fopen( szModuleName, "rb" );
    if ( !pFile )
        return FALSE;

    // Seek to the NE header
    fseek( pFile, 0x3C, SEEK_SET );
    fread( &neOffset, sizeof(neOffset), 1, pFile );
    fseek( pFile, neOffset, SEEK_SET );
    fread( &neheader, sizeof(neheader), 1, pFile );
    if ( neheader.signature != 0x454E ) // "NE"
    {
        fclose( pFile );
        return FALSE;
    }

    // Iterate through each of the segments...
    for ( i=0; i < neheader.segment_count; i++ )
    {
        SEGMENT_RECORD segRec;
        WORD reloc_count;
        COMMON_RELOCATION_RECORD far * lpRelocs;
        WORD j;

        // Seek to and read the i'th segment's entry in the seg table
        fseek( pFile,
               neOffset + neheader.segment_offset + (i*sizeof(SEGMENT_RECORD)),
               SEEK_SET );
        fread( &segRec, sizeof(segRec), 1, pFile );

        if ( (!segRec.reloc_info) || (!segRec.sector_offset) )
            continue;

        // Seek to the start of the segment's relocation table.  This is
        // immediately after the raw segment data in the file.
        fseek ( pFile,
                (DWORD)segRec.sector_offset * (1 << neheader.shift_count)
                    + segRec.segment_length,
                SEEK_SET );

        fread( &reloc_count, sizeof(reloc_count), 1, pFile );
        if ( reloc_count <= 1 )
            continue;

        // Read in the number of relocations (a WORD).  Limit to 8192
        reloc_count = min( reloc_count, 8192 );
        lpRelocs =
            (COMMON_RELOCATION_RECORD far *)GlobalAllocPtr(GMEM_MOVEABLE,
                          reloc_count * sizeof(COMMON_RELOCATION_RECORD));
        if ( !lpRelocs )
            continue;

        // Read in the relocations as an array
        fread(lpRelocs, sizeof(COMMON_RELOCATION_RECORD), reloc_count, pFile);

        // Sort the relocations so we can find duplicates easily.
        qsort( lpRelocs, reloc_count, sizeof(COMMON_RELOCATION_RECORD),
               SortRelocs );

        // Go through all the relocations, comparing successive records.
        // If they match (other than the 2nd WORD of the record), they're
        // duplicates, and could be chained.
        for ( j = 0; j < reloc_count-1; j++ )
            if ( (lpRelocs[j].types == lpRelocs[j+1].types)
              && (lpRelocs[j].variable_data == lpRelocs[j+1].variable_data) )
            {
                if ( (lpRelocs[j].types & 0x0400) == 0 )  // Don't count
                    cDuplicates++;                        // additive fixups
            }

        // Done with this segment's relocation data.  Free up the memory.
        GlobalFreePtr( lpRelocs );
    }

    fclose( pFile );
    return cDuplicates;
}

// qsort() sorting function for NE fixups.  Sort first by the fixup's
// first WORD, then by the final DWORD.  Note that the WORD and DWORD
// were sorting by are composed of BYTE & WORD fields.  We're just choosing
// to combine them for speed of sorting/searching.
int SortRelocs(const void * a, const void *b)
{
    LONG result;
    WORD a_types, b_types;
    DWORD a_variable_data, b_variable_data;

    a_types = ((COMMON_RELOCATION_RECORD far *)a)->types;
    b_types = ((COMMON_RELOCATION_RECORD far *)b)->types;

    if ( a_types != b_types )
        return (int) (a_types - b_types);

    a_variable_data = ((COMMON_RELOCATION_RECORD far *)a)->variable_data;
    b_variable_data = ((COMMON_RELOCATION_RECORD far *)b)->variable_data;

    result = (LONG) (a_variable_data - b_variable_data);

     if ( result < 0 )
         return -1;
     return result ? 1 : 0;
}
